# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs


import os
import shutil

import docker
import requests

from observatory.platform.docker.compose_runner import ComposeRunner
from observatory.platform.observatory_config import Config
from observatory.platform.utils.config_utils import module_file_path

HOST_UID = os.getuid()
DEBUG = False


class PlatformRunner(ComposeRunner):
    def __init__(self, *, config: Config, host_uid: int = HOST_UID, docker_build_path: str = None, debug: bool = DEBUG):
        """Create a PlatformRunner instance, which is used to build, start and stop an Observatory Platform instance.

        :param config: the config.
        :param host_uid: The user id of the host system. Used to set the user id in the Docker containers.
        :param docker_build_path: the Docker build path.
        :param debug: Print debugging information.
        """

        self.config = config
        self.host_uid = host_uid

        # Set default values when config is invalid
        observatory_home = self.config.observatory.observatory_home

        if docker_build_path is None:
            docker_build_path = os.path.join(observatory_home, "build", "docker")

        super().__init__(
            compose_template_path=os.path.join(self.docker_module_path, "docker-compose.observatory.yml.jinja2"),
            build_path=docker_build_path,
            compose_template_kwargs={"config": self.config},
            debug=debug,
        )

        # Add files
        self.add_template(
            path=os.path.join(self.docker_module_path, "Dockerfile.observatory.jinja2"), config=self.config
        )
        self.add_template(
            path=os.path.join(self.docker_module_path, "entrypoint-airflow.sh.jinja2"), config=self.config
        )
        self.add_template(
            path=os.path.join(self.docker_module_path, "Dockerfile.apiserver.jinja2"), config=self.config
        )
        self.add_template(
            path=os.path.join(self.docker_module_path, "Dockerfile.seed_db.jinja2"), config=self.config
        )
        self.add_template(
            path=os.path.join(self.docker_module_path, "entrypoint-api.sh.jinja2"), config=self.config
        )
        self.add_file(
            path=os.path.join(self.docker_module_path, "entrypoint-root.sh"), output_file_name="entrypoint-root.sh"
        )
        self.add_template(path=os.path.join(self.docker_module_path, "seed_db.sh.jinja2"), config=self.config)

        # Add all project requirements files for local projects
        for package in self.config.python_packages:
            if package.type == "editable":
                # Add requirements.sh
                self.add_file(
                    path=os.path.join(package.host_package, "requirements.sh"),
                    output_file_name=f"requirements.{package.name}.sh",
                )
            elif package.type == "sdist":
                # Add sdist package file
                self.add_file(path=package.host_package, output_file_name=package.docker_package)

    @property
    def is_environment_valid(self) -> bool:
        """Return whether the environment for building the Observatory Platform is valid.

        :return: whether the environment for building the Observatory Platform is valid.
        """

        return all([self.docker_exe_path is not None, self.docker_compose_path is not None, self.is_docker_running])

    @property
    def docker_module_path(self) -> str:
        """The path to the Observatory Platform docker module.

        :return: the path.
        """

        return module_file_path("observatory.platform.docker")

    @property
    def docker_exe_path(self) -> str:
        """The path to the Docker executable.

        :return: the path or None.
        """

        return shutil.which("docker")

    @property
    def docker_compose_path(self) -> str:
        """The path to the Docker Compose executable.

        :return: the path or None.
        """

        return shutil.which("docker-compose")

    @property
    def is_docker_running(self) -> bool:
        """Checks whether Docker is running or not.

        :return: whether Docker is running or not.
        """

        client = docker.from_env()
        try:
            is_running = client.ping()
        except requests.exceptions.ConnectionError:
            is_running = False
        return is_running

    def make_environment(self):
        """Make an environment containing the environment variables that are required to build and start the
        Observatory docker environment.

        :return: None.
        """

        env = os.environ.copy()

        # Sets the name
        env["COMPOSE_PROJECT_NAME"] = self.config.observatory.docker_compose_project_name

        # Host settings
        env["HOST_USER_ID"] = str(self.host_uid)
        env["HOST_OBSERVATORY_HOME"] = os.path.normpath(self.config.observatory.observatory_home)
        env["HOST_REDIS_PORT"] = str(self.config.observatory.redis_port)
        env["HOST_FLOWER_UI_PORT"] = str(self.config.observatory.flower_ui_port)
        env["HOST_AIRFLOW_UI_PORT"] = str(self.config.observatory.airflow_ui_port)
        env["HOST_ELASTIC_PORT"] = str(self.config.observatory.elastic_port)
        env["HOST_KIBANA_PORT"] = str(self.config.observatory.kibana_port)
        env["HOST_API_PORT"] = str(self.config.observatory.api_port)

        # Secrets
        if self.config.google_cloud is not None and self.config.google_cloud.credentials is not None:
            env["HOST_GOOGLE_APPLICATION_CREDENTIALS"] = self.config.google_cloud.credentials
        env["AIRFLOW_FERNET_KEY"] = self.config.observatory.airflow_fernet_key
        env["AIRFLOW_SECRET_KEY"] = self.config.observatory.airflow_secret_key
        env["AIRFLOW_UI_USER_EMAIL"] = self.config.observatory.airflow_ui_user_email
        env["AIRFLOW_UI_USER_PASSWORD"] = self.config.observatory.airflow_ui_user_password
        env["POSTGRES_PASSWORD"] = self.config.observatory.postgres_password

        # Create Airflow variables
        airflow_variables = self.config.make_airflow_variables()
        for variable in airflow_variables:
            env[variable.env_var_name] = str(variable.value)

        # Airflow connections
        for conn in self.config.airflow_connections:
            env[conn.conn_name] = conn.value

        return env

    def make_files(self):
        """Create directories that are mounted as volumes as defined in the docker-compose file.

        :return: None.
        """
        super(PlatformRunner, self).make_files()
        observatory_home = os.path.normpath(self.config.observatory.observatory_home)
        # Create data directory
        data_dir = os.path.join(observatory_home, "data")
        os.makedirs(data_dir, exist_ok=True)

        # Create logs directory
        logs_dir = os.path.join(observatory_home, "logs")
        os.makedirs(logs_dir, exist_ok=True)

        # Create postgres directory
        postgres_dir = os.path.join(observatory_home, "postgres")
        os.makedirs(postgres_dir, exist_ok=True)



# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import dataclasses
import os
import subprocess
from subprocess import Popen
from typing import Dict, List

from observatory.platform.docker.builder import Builder, rendered_file_name
from observatory.platform.utils.proc_utils import stream_process


@dataclasses.dataclass
class ProcessOutput:
    """Output from a process

    :param output: the process std out.
    :param error: the process std error.
    :param return_code: the process return code.
    """

    output: str
    error: str
    return_code: int


class ComposeRunner(Builder):
    COMPOSE_ARGS_PREFIX = ["docker-compose", "-f"]
    COMPOSE_BUILD_ARGS = ["build"]
    COMPOSE_START_ARGS = ["up", "-d"]
    COMPOSE_STOP_ARGS = ["down"]

    def __init__(
        self, *, compose_template_path: str, build_path: str, compose_template_kwargs: Dict = None, debug: bool = False
    ):
        """Docker Compose runner constructor.

        :param compose_template_path: the path to the Docker Compose Jinja2 template file.
        :param build_path: the path where Docker will build.
        :param compose_template_kwargs: the kwargs to use when rendering the Docker Compose Jinja2 template file.
        :param debug: whether to run in debug mode or not.
        """

        super().__init__(build_path=build_path)
        if compose_template_kwargs is None:
            compose_template_kwargs = dict()

        self.debug = debug
        self.compose_template_path = compose_template_path
        self.add_template(path=compose_template_path, **compose_template_kwargs)

    def make_environment(self) -> Dict:
        """Make the environment variables.

        :return: environment dictionary.
        """

        return os.environ.copy()

    @property
    def compose_file_name(self):
        """Return the file name for the rendered Docker Compose template.

        :return: Docker Compose file name.
        """

        return rendered_file_name(self.compose_template_path)

    def build(self) -> ProcessOutput:
        """Build the Docker containers.

        :return: output and error stream results and proc return code.
        """

        return self.__run_docker_compose_cmd(self.COMPOSE_BUILD_ARGS)

    def start(self) -> ProcessOutput:
        """Start the Docker containers.

        :return: output and error stream results and proc return code.
        """

        return self.__run_docker_compose_cmd(self.COMPOSE_START_ARGS)

    def stop(self) -> ProcessOutput:
        """Stop the Docker containers.

        :return: output and error stream results and proc return code.
        """

        return self.__run_docker_compose_cmd(self.COMPOSE_STOP_ARGS)

    def __run_docker_compose_cmd(self, args: List) -> ProcessOutput:
        """Run a set of Docker Compose arguments.

        :param args: the list of arguments.
        :return: output and error stream results and proc return code.
        """

        # Make environment
        env = self.make_environment()

        # Make files
        self.make_files()

        # Build the containers first
        proc: Popen = subprocess.Popen(
            self.COMPOSE_ARGS_PREFIX + [self.compose_file_name] + args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
            cwd=self.build_path,
        )

        # Wait for results
        output, error = stream_process(proc, self.debug)

        return ProcessOutput(output, error, proc.returncode)


# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose

from __future__ import annotations

import json
import logging
import os
from dataclasses import dataclass
from enum import Enum
from http import HTTPStatus
from typing import Tuple, List

import requests


class TerraformVariableCategory(Enum):
    """The type of Terraform variable."""

    terraform = "terraform"
    env = "env"


@dataclass
class TerraformVariable:
    """A TerraformVariable class.

    Attributes:
        key: variable key.
        value: variable value.
        var_id: variable id, uniquely identifies the variable in the cloud.
        description: variable description.
        category: variable category.
        hcl: whether the variable is HCL or not.
        sensitive: whether the variable is sensitive or not.
    """

    key: str
    value: str
    var_id: str = None
    description: str = ""
    category: TerraformVariableCategory = TerraformVariableCategory.terraform
    hcl: bool = False
    sensitive: bool = False

    def __str__(self):
        return self.key

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return self.key == other.key

    @staticmethod
    def from_dict(dict_) -> TerraformVariable:
        """Parse a dictionary into a TerraformVariable instance.

        :param dict_: the dictionary.
        :return: the TerraformVariable instance.
        """

        var_id = dict_.get("id")
        attributes = dict_["attributes"]
        key = attributes.get("key")
        value = attributes.get("value")
        sensitive = attributes.get("sensitive")
        category = attributes.get("category")
        hcl = attributes.get("hcl")
        description = attributes.get("description")

        return TerraformVariable(
            key,
            value,
            sensitive=sensitive,
            category=TerraformVariableCategory(category),
            hcl=hcl,
            description=description,
            var_id=var_id,
        )

    def to_dict(self):
        """Convert a TerraformVariable instance into a dictionary.

        :return: the dictionary.
        """

        var = {
            "type": "vars",
            "attributes": {
                "key": self.key,
                "value": self.value,
                "description": self.description,
                "category": self.category.value,
                "hcl": self.hcl,
                "sensitive": self.sensitive,
            },
        }

        if self.var_id is not None:
            var["id"] = self.var_id

        return var


class TerraformApi:
    TERRAFORM_WORKSPACE_VERSION = "0.13.5"
    VERBOSITY_WARNING = 0
    VERBOSITY_INFO = 1
    VERBOSITY_DEBUG = 2

    def __init__(self, token: str, verbosity: int = VERBOSITY_WARNING):
        """Create a TerraformApi instance.

        :param token: the Terraform API token.
        :param verbosity: the verbosity for the Terraform API.
        """

        self.token = token
        if verbosity == TerraformApi.VERBOSITY_WARNING:
            logging.getLogger().setLevel(logging.WARNING)
        elif verbosity == TerraformApi.VERBOSITY_INFO:
            logging.getLogger().setLevel(logging.INFO)
        elif verbosity >= TerraformApi.VERBOSITY_DEBUG:
            logging.getLogger().setLevel(logging.DEBUG)
        self.api_url = "https://app.terraform.io/api/v2"
        self.headers = {"Content-Type": "application/vnd.api+json", "Authorization": f"Bearer {token}"}

    @staticmethod
    def token_from_file(file_path: str) -> str:
        """Get the Terraform token from a credentials file.

        :param file_path: path to credentials file
        :return: token
        """

        with open(file_path, "r") as file:
            token = json.load(file)["credentials"]["app.terraform.io"]["token"]
        return token

    def create_workspace(
        self,
        organisation: str,
        workspace: str,
        auto_apply: bool,
        description: str,
        version: str = TERRAFORM_WORKSPACE_VERSION,
    ) -> int:
        """Create a new workspace in Terraform Cloud.

        :param organisation: Name of terraform organisation
        :param workspace: Name of terraform workspace
        :param auto_apply: Whether the new workspace should be set to auto_apply
        :param description: Description of the workspace
        :param version: The terraform version
        :return: The response status code
        """
        attributes = {
            "name": workspace,
            "auto-apply": str(auto_apply).lower(),
            "description": description,
            "terraform_version": version,
        }
        data = {"data": {"type": "workspaces", "attributes": attributes}}

        response = requests.post(
            f"{self.api_url}/organizations/{organisation}/workspaces", headers=self.headers, json=data
        )

        if response.status_code == HTTPStatus.CREATED:
            logging.info(f"Created workspace {workspace}")
            logging.debug(f"response: {response.text}")
        elif response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            logging.warning(f"Workspace with name {workspace} already exists")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful creating workspace, response: {response.text}")
            exit(os.EX_CONFIG)
        return response.status_code

    def delete_workspace(self, organisation: str, workspace: str) -> int:
        """Delete a workspace in terraform cloud.

        :param organisation: Name of terraform organisation
        :param workspace: Name of terraform workspace
        :return: The response status code
        """

        response = requests.delete(
            f"{self.api_url}/organizations/{organisation}/workspaces/{workspace}", headers=self.headers
        )
        return response.status_code

    def workspace_id(self, organisation: str, workspace: str) -> str:
        """Returns the workspace id.

        :param organisation: Name of terraform organisation
        :param workspace: Name of terraform workspace
        :return: workspace id
        """
        response = requests.get(
            f"{self.api_url}/organizations/{organisation}/workspaces/{workspace}", headers=self.headers
        )
        if response.status_code == HTTPStatus.OK:
            logging.info(f"Retrieved workspace id for workspace '{workspace}'.")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(
                f"Unsuccessful retrieving workspace id for workspace '{workspace}', response: {response.text}"
            )
            exit(os.EX_CONFIG)

        workspace_id = json.loads(response.text)["data"]["id"]
        return workspace_id

    def add_workspace_variable(self, variable: TerraformVariable, workspace_id: str) -> str:
        """Add a new variable to the workspace. Will return an error if the variable already exists.

        :param variable: the TerraformVariable instance.
        :param workspace_id: the workspace id
        :return: The var id
        """

        response = requests.post(
            f"{self.api_url}/workspaces/{workspace_id}/vars", headers=self.headers, json={"data": variable.to_dict()}
        )

        key = variable.key
        if response.status_code == HTTPStatus.CREATED:
            logging.info(f"Added variable {key}")
        else:
            msg = f"Unsuccessful adding variable {key}, response: {response.text}, status_code: {response.status_code}"
            logging.error(msg)
            raise ValueError(msg)

        var_id = json.loads(response.text)["data"]["id"]
        return var_id

    def update_workspace_variable(self, variable: TerraformVariable, workspace_id: str) -> int:
        """Update a workspace variable that is identified by its id.

        :param variable: attributes of the variable
        :param var_id: the variable id
        :param workspace_id: the workspace id
        :return: the response status code
        """
        response = requests.patch(
            f"{self.api_url}/workspaces/{workspace_id}/vars/{variable.var_id}",
            headers=self.headers,
            json={"data": variable.to_dict()},
        )
        try:
            key = json.loads(response.text)["data"]["attributes"]["key"]
        except KeyError:
            try:
                key = variable.key
            except KeyError:
                key = None
        if response.status_code == HTTPStatus.OK:
            logging.info(f"Updated variable {key}")
        else:
            msg = f"Unsuccessful updating variable with id {variable.var_id} and key {key}, response: {response.text}, status_code: {response.status_code}"
            logging.error(msg)
            raise ValueError(msg)

        return response.status_code

    def delete_workspace_variable(self, var: TerraformVariable, workspace_id: str) -> int:
        """Delete a workspace variable identified by its id. Should not return any content in response.

        :param var: the variable
        :param workspace_id: the workspace id
        :return: The response code
        """
        response = requests.delete(f"{self.api_url}/workspaces/{workspace_id}/vars/{var.var_id}", headers=self.headers)

        if response.status_code == HTTPStatus.NO_CONTENT:
            logging.info(f"Deleted variable with id {var.var_id}")
        else:
            msg = f"Unsuccessful deleting variable with id {var.var_id}, response: {response.text}, status_code: {response.status_code}"
            logging.error(msg)
            raise ValueError(msg)

        return response.status_code

    def list_workspace_variables(self, workspace_id: str) -> List[TerraformVariable]:
        """Returns a list of variables in the workspace. Each variable is a dict.

        :param workspace_id: The workspace id
        :return: Variables in the workspace
        """
        response = requests.get(f"{self.api_url}/workspaces/{workspace_id}/vars", headers=self.headers)

        if response.status_code == HTTPStatus.OK:
            logging.info(f"Retrieved workspace variables.")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful retrieving workspace variables, response: {response.text}")
            exit(os.EX_CONFIG)

        workspace_vars = json.loads(response.text)["data"]
        return [TerraformVariable.from_dict(dict_) for dict_ in workspace_vars]

    def create_configuration_version(self, workspace_id: str) -> Tuple[str, str]:
        """Create a configuration version. A configuration version is a resource used to reference the uploaded
        configuration files. It is associated with the run to use the uploaded configuration files for performing the
        plan and apply.

        :param workspace_id: the workspace id
        :return: the upload url
        """
        data = {"data": {"type": "configuration-versions", "attributes": {"auto-queue-runs": "false"}}}
        response = requests.post(
            f"{self.api_url}/workspaces/{workspace_id}/configuration-versions", headers=self.headers, json=data
        )
        if response.status_code == 201:
            logging.info(f"Created configuration version.")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful creating configuration version, response: {response.text}")
            exit(os.EX_CONFIG)

        upload_url = json.loads(response.text)["data"]["attributes"]["upload-url"]
        configuration_id = json.loads(response.text)["data"]["id"]
        return upload_url, configuration_id

    def get_configuration_version_status(self, configuration_id: str) -> str:
        """Show the configuration version and return it's status. The status will be pending when the
        configuration version is initially created and will remain pending until configuration files are supplied via
        upload, and while they are processed. The status will then be changed to 'uploaded'. Runs cannot be created
        using pending or errored configuration versions.

        :param configuration_id: the configuration version id
        :return: configuration version status
        """
        response = requests.get(f"{self.api_url}/configuration-versions/{configuration_id}", headers=self.headers)
        if response.status_code == HTTPStatus.OK:
            logging.info(f"Retrieved configuration version info.")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful retrieving configuration version info, response: {response.text}")
            exit(os.EX_CONFIG)

        status = json.loads(response.text)["data"]["attributes"]["status"]
        return status

    @staticmethod
    def upload_configuration_files(upload_url: str, configuration_path: str) -> int:
        """Uploads the configuration files. Auto-queue-runs is set to false when creating configuration version,
        so conf will not be queued automatically.

        :param upload_url: upload url, returned when creating configuration version
        :param configuration_path: path to tar.gz file containing config files (main.tf)
        :return: the response code
        """

        headers = {"Content-Type": "application/octet-stream"}
        with open(configuration_path, "rb") as configuration:
            response = requests.put(upload_url, headers=headers, data=configuration.read())
        if response.status_code == HTTPStatus.OK:
            logging.info(f"Uploaded configuration.")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful uploading configuration, response: {response.text}")
            exit(os.EX_CONFIG)

        return response.status_code

    def create_run(self, workspace_id: str, target_addrs: str = None, message: str = "") -> str:
        """Creates a run, optionally targeted at a target address. If auto-apply is set to true the run will be applied
        afterwards as well.

        :param workspace_id: the workspace id
        :param target_addrs: the target address (id of the module/resource)
        :param message: additional message that will be displayed at the terraform cloud run
        :return: the run id
        """

        data = {
            "data": {
                "attributes": {"message": message},
                "type": "runs",
                "relationships": {"workspace": {"data": {"type": "workspaces", "id": workspace_id}}},
            }
        }
        if target_addrs:
            data["data"]["attributes"]["target-addrs"] = [target_addrs]

        response = requests.post(f"{self.api_url}/runs", headers=self.headers, json=data)
        if response.status_code == HTTPStatus.CREATED:
            logging.info(f"Created run.")
            logging.debug(f"response: {response.text}")
        else:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful creating run, response: {response.text}")
            exit(os.EX_CONFIG)

        run_id = json.loads(response.text)["data"]["id"]
        return run_id

    def get_run_details(self, run_id: str) -> dict:
        """Get details on a run identified by its id.
        :param run_id: the run id
        :return: the response text
        """

        response = requests.get(f"{self.api_url}/runs/{run_id}", headers=self.headers)
        if not response.status_code == HTTPStatus.OK:
            logging.error(f"Response status: {response.status_code}")
            logging.error(f"Unsuccessful retrieving run details, response: {response.text}")
            exit(os.EX_CONFIG)

        return json.loads(response.text)

    def plan_variable_changes(
        self, new_vars: List[TerraformVariable], workspace_id: str
    ) -> Tuple[
        List[TerraformVariable],
        List[Tuple[TerraformVariable, TerraformVariable]],
        List[TerraformVariable],
        List[TerraformVariable],
    ]:
        """Compares the current variables in the workspace with a list of new variables. It sorts the new variables in
        one of 4 different categories and adds them to the corresponding list. Sensitive variables can never be
        'unchanged'.

        :param new_vars: list of potential new variables where each variable is a variable attributes dict
        :param workspace_id: the workspace id
        :return: lists of variables in different categories (add, edit, unchanged, delete).
        add: list of attributes dicts
        edit: list of tuples with (attributes dict, var id, old value)
        unchanged: list of attributes dicts
        delete: list of tuples with (var key, var id, old value)
        """

        add: List[TerraformVariable] = []
        edit: List[Tuple[TerraformVariable, TerraformVariable]] = []
        unchanged: List[TerraformVariable] = []

        # Get existing variables
        old_vars = self.list_workspace_variables(workspace_id)

        # Make dict with old variable keys as keys
        old_var_ids = {}
        for old_var in old_vars:
            old_var_ids[old_var.key] = old_var

        # Check which new variables need to be updated and which need to be added
        for new_var in new_vars:
            # check if variable already exists
            if new_var.key in old_var_ids.keys():
                old_var = old_var_ids[new_var.key]

                # check if values of old and new are the same
                if new_var.sensitive or new_var.value != old_var.value:
                    # Assign id of old variable to new variable
                    new_var.var_id = old_var.var_id
                    edit.append((old_var, new_var))
                else:
                    unchanged.append(new_var)
            else:
                add.append(new_var)

        # Compare old variables with new variables and check which are 'extra' in the old vars
        delete = list(set(old_vars) - set(new_vars))

        return add, edit, unchanged, delete

    def update_workspace_variables(self, add: list, edit: list, delete: list, workspace_id: str):
        """Update workspace accordingly to the planned changes.

        :param add: list of attributes dicts of new variables
        :param edit: list of tuples with (attributes dict, var id, old value) of existing variables that will be updated
        :param delete: list of tuples with (var key, var id, old value) of variables that will be deleted
        :param workspace_id: the workspace id
        :return: None
        """

        for var in add:
            self.add_workspace_variable(var, workspace_id)
        for old_var, new_var in edit:
            self.update_workspace_variable(new_var, workspace_id)
        for var in delete:
            self.delete_workspace_variable(var, workspace_id)



# Copyright 2020 Artificial Dimensions Ltd
# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

from jinja2 import Environment, FileSystemLoader
import os


def render_template(template_path: str, **kwargs) -> str:
    """Render a Jinja2 template.

    :param template_path: the path to the template.
    :param kwargs: the keyword variables to populate the template with.
    :return: the rendered template as a string.
    """

    # Read file contents
    with open(template_path, "r") as file:
        contents = file.read()

    dir = os.path.dirname(template_path)
    template = Environment(loader=FileSystemLoader(dir)).from_string(contents)
    rendered = template.render(**kwargs)

    return rendered


def make_jinja2_filename(file_name: str) -> str:
    """Add .jinja2 to a filename.

    :param file_name: the filename without an extension.
    :return: the filename.
    """
    return f"{file_name}.jinja2"


def make_sql_jinja2_filename(file_name: str) -> str:
    """Add .sql.jinja2 to a filename.

    :param file_name: the filename without an extension.
    :return: the filename.
    """
    return f"{file_name}.sql.jinja2"



# Copyright 2021 Curtin University. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien, Aniek Roelofs

import json
import os
import unittest

from click.testing import CliRunner

from docs.generate_schema_csv import generate_csv, generate_latest_files, schema_to_csv


class TestSchemaCSVGenerator(unittest.TestCase):

    test_schema = [
        {"name": "test", "type": "test_type", "mode": "test_mode", "description": "test_description"},
        {
            "name": "test2",
            "type": "RECORD",
            "mode": "test_mode",
            "description": "test_description",
            "fields": [{"name": "test3", "type": "test_type", "mode": "test_mode"}],
        },
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_schema_to_csv_prefix(self):
        rows = list()
        schema_to_csv(schema=self.test_schema, output=rows)
        self.assertEqual(len(rows), 3)
        self.assertEqual(rows[0]["name"], "test")
        self.assertEqual(rows[0]["type"], "test_type")
        self.assertEqual(rows[0]["mode"], "test_mode")
        self.assertEqual(rows[0]["description"], "test_description")

        self.assertEqual(rows[1]["name"], "test2")
        self.assertEqual(rows[1]["type"], "RECORD")
        self.assertEqual(rows[1]["mode"], "test_mode")
        self.assertEqual(rows[1]["description"], "test_description")

        self.assertEqual(rows[2]["name"], "test2.test3")
        self.assertEqual(rows[2]["type"], "test_type")
        self.assertEqual(rows[2]["mode"], "test_mode")
        self.assertEqual(rows[2]["description"], None)

    def test_generate_csv(self):
        with CliRunner().isolated_filesystem():
            schemas_dir = "test_schemas"
            os.makedirs(schemas_dir, exist_ok=True)
            with open(os.path.join(schemas_dir, "test_schema_2021-01-01.json"), "w") as f:
                f.write(json.dumps(self.test_schema, indent=4))
            generate_csv(schema_dir=os.path.join(os.getcwd(), schemas_dir))
            self.assertTrue(os.path.exists(os.path.join("schemas", "test_schema_2021-01-01.csv")))

    def test_generate_latest_files(self):
        with CliRunner().isolated_filesystem():
            schemas_dir = "test_schemas"
            os.makedirs(schemas_dir, exist_ok=True)
            with open(os.path.join(schemas_dir, "test_schema_2021-01-01.json"), "w") as f:
                f.write(json.dumps(self.test_schema, indent=4))
            generate_csv(schema_dir=os.path.join(os.getcwd(), schemas_dir))
            generate_latest_files()
            self.assertTrue(os.path.exists(os.path.join("schemas", "test_schema_latest.csv")))



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs

""" General google cloud utility functions (independent of telescope usage) """

import json
import logging
import multiprocessing
import os
import re
import time
from concurrent.futures import ProcessPoolExecutor, as_completed
from copy import deepcopy
from enum import Enum
from multiprocessing import BoundedSemaphore, cpu_count
from typing import List, Optional, Tuple, Union, Any

import pendulum
from airflow.models import Variable
from google.api_core.exceptions import BadRequest, Conflict
from google.cloud import bigquery, storage
from google.cloud.bigquery import LoadJob, LoadJobConfig, QueryJob, SourceFormat, CopyJobConfig, CopyJob
from google.cloud.bigquery.job import QueryJobConfig
from google.cloud.exceptions import Conflict, NotFound
from google.cloud.storage import Blob
from googleapiclient import discovery as gcp_api
from observatory.api.client.model.big_query_bytes_processed import (
    BigQueryBytesProcessed,
)
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.api import make_observatory_api
from observatory.platform.utils.config_utils import utils_templates_path
from observatory.platform.utils.file_utils import crc32c_base64_hash
from observatory.platform.utils.jinja2_utils import (
    make_sql_jinja2_filename,
    render_template,
)
from requests.exceptions import ChunkedEncodingError

# The chunk size to use when uploading / downloading a blob in multiple parts, must be a multiple of 256 KB.
DEFAULT_CHUNK_SIZE = 256 * 1024 * 4

# BigQuery daily query byte limit.
BIGQUERY_QUERY_DAILY_BYTE_LIMIT = 1024 * 1024 * 1024 * 1024 * 5  # 5 TiB


# This function makes it easier to mock out in tests. This mechanism is here to allow you to hard code disabling the
# daily quota check across the platform.  When you are convinced the feature is working correctly in production and
# that you wan't it always on, then the mechanism can be removed.
def bq_query_daily_limit_enabled():
    """Whether to turn on BigQuery daily limits.

    :return: Whether to use daily limits.
    """

    return True


def bq_query_bytes_estimate(*args, **kwargs) -> int:
    """Do a dry run of a BigQuery query to estimate the bytes processed.

    :param args: Positional arguments to pass onto the callable.
    :param kwargs: Named arguments to pass onto the callable.
    :return: Query bytes estimate.
    """

    if "job_config" not in kwargs:
        kwargs["job_config"] = QueryJobConfig()

    config = deepcopy(kwargs["job_config"])
    config.dry_run = True
    kwargs["job_config"] = config

    bytes_estimate = bigquery.Client().query(*args, **kwargs).total_bytes_processed
    return bytes_estimate


def bq_query_bytes_budget_check(*, bytes_budget: Optional[int], bytes_estimate: int):
    """Check that the estimated number of processed bytes required does not exceed the budgeted number of bytes for the
    query. If the estimate exceeds the budget, this function throws an exception.

    :param bytes_budget: The processed bytes budget for this query.
    :param bytes_estimate: Estimated number of bytes processed in query.
    """

    if bytes_budget is None:
        return

    if bytes_estimate > bytes_budget:
        raise Exception(f"Bytes estimate {bytes_estimate} exceeds the budget {bytes_budget}.")


def get_bytes_processed(*, api, project: str) -> int:
    """Get the bytes processed over the last 24 hours.
    :param api: Observatory platform API client object.
    :param project: GCP project.
    :return: Number of bytes processed.
    """

    return api.get_bigquery_bytes_processed(project=project)


def save_bytes_processed(*, api: Any, project: str, bytes_estimate: int):
    """Update the number of bytes processed.
    :param api: Observatory platform API client object.
    :param project: GCP project.
    :param bytes_estimate: Bytes estimate of the current query.
    """

    record = BigQueryBytesProcessed(
        project=project,
        total=bytes_estimate,
    )
    api.post_bigquery_bytes_processed(record)


def bq_query_bytes_daily_limit_check(bytes_estimate: int) -> BigQueryBytesProcessed:
    """Check the daily byte limit. Raise exception if the current query will put us over the limit.

    :param bytes_estimate: Estimated number of bytes processed in query.
    """

    # Disable if other repos not ready yet
    if not bq_query_daily_limit_enabled():
        return

    try:
        project = Variable.get(AirflowVars.PROJECT_ID)
        api = make_observatory_api()
    except Exception as e:
        logging.warning(f"Skipping daily byte limit check: {e}")
        return

    # Get bytes processed over last 24 hours
    bytes_processed = get_bytes_processed(api=api, project=project)

    # Check that the bytes processed over the last 24 hours + the estimated query amount is below the daily limit
    projected_estimate = bytes_processed + bytes_estimate
    if projected_estimate > BIGQUERY_QUERY_DAILY_BYTE_LIMIT:
        raise Exception(
            f"The projected bytes estimate of {projected_estimate} exceeds the daily limit of {BIGQUERY_QUERY_DAILY_BYTE_LIMIT}"
        )

    # Save the estimate, assumes the actual query will run after
    save_bytes_processed(api=api, project=project, bytes_estimate=bytes_estimate)


def table_name_from_blob(blob_name: str, file_extension: str):
    """Make a BigQuery table name from a blob name.

    :param blob_name: the blob name.
    :param file_extension: the file extension of the blob.
    :return: the table name.
    """

    assert "." in file_extension, "file_extension must contain a ."
    file_name = os.path.basename(blob_name)
    match = re.match(rf".+?(?={file_extension})", file_name)
    if match is None:
        raise ValueError(f"Could not find table name from blob_name={blob_name}")
    return match.group(0)


class TransferStatus(Enum):
    """The status of the Google Cloud Data Transfer operation"""

    in_progress = "IN_PROGRESS"
    success = "SUCCESS"
    aborted = "ABORTED"
    failed = "FAILED"


def bigquery_table_exists(project_id: str, dataset_id: str, table_name: str) -> bool:
    """Checks whether a BigQuery table exists or not.

    :param project_id: the Google Cloud project id.
    :param dataset_id: the BigQuery dataset id.
    :param table_name: the name of the table.
    :return: whether the table exists or not.
    """

    client = bigquery.Client(project_id)
    dataset = bigquery.Dataset(f"{project_id}.{dataset_id}")
    table = dataset.table(table_name)
    table_exists = True

    try:
        client.get_table(table)
    except NotFound:
        table_exists = False

    return table_exists


def bigquery_sharded_table_id(table_name, datetime: pendulum.Date) -> str:
    """Create a sharded table identifier for a BigQuery table.

    :param table_name: the name of the table.
    :param datetime: the date to append as a shard suffix.
    :return: the table id.
    """
    return f"{table_name}{datetime.strftime('%Y%m%d')}"


def create_bigquery_dataset(project_id: str, dataset_id: str, location: str, description: str = "") -> bigquery.Dataset:
    """Create a BigQuery dataset.

    :param project_id: the Google Cloud project id.
    :param dataset_id: the BigQuery dataset id.
    :param location: the location where the dataset will be stored:
    https://cloud.google.com/compute/docs/regions-zones/#locations
    :param description: a description for the dataset
    :return: None
    """

    func_name = create_bigquery_dataset.__name__

    # Make the dataset reference
    dataset_ref = f"{project_id}.{dataset_id}"

    # Make dataset handle
    client = bigquery.Client()
    dataset = bigquery.Dataset(dataset_ref)

    # Set properties
    dataset.location = location
    dataset.description = description

    # Create dataset, if already exists then catch exception
    try:
        logging.info(f"{func_name}: creating dataset dataset_ref={dataset_ref}")
        dataset = client.create_dataset(dataset)
    except Conflict as e:
        logging.warning(f"{func_name}: dataset already exists dataset_ref={dataset_ref}, exception={e}")
    return dataset


def create_empty_bigquery_table(
    dataset_id: str,
    location: str,
    table_id: str,
    schema_file_path: str = None,
    project_id: str = None,
    clustering_fields: List = None,
):
    """Creates an empty BigQuery table. If a path to a schema file is given the table will be created using this
    schema.

    :param dataset_id: BigQuery dataset id.
    :param location: location of the BigQuery dataset.
    :param table_id: BigQuery table name.
    :param schema_file_path: path on local file system to BigQuery table schema.
    :param project_id: Google Cloud project id.
    :param clustering_fields: what fields to cluster on.
    :return: The table instance if the request was successfull.
    """
    func_name = create_empty_bigquery_table.__name__
    msg = f"dataset_id={dataset_id}, location={location}, table={table_id}, " f"schema_file_path={schema_file_path}"
    logging.info(f"{func_name}: creating empty bigquery table {msg}")

    client = bigquery.Client()
    if project_id is None:
        project_id = client.project

    dataset = create_bigquery_dataset(project_id, dataset_id, location)

    if schema_file_path:
        schema = client.schema_from_json(schema_file_path)
        table = bigquery.Table(dataset.table(table_id), schema=schema)
    else:
        table = bigquery.Table(dataset.table(table_id))

    # Note that clustering fields can only be set if a schema is specified
    if clustering_fields:
        table.clustering_fields = clustering_fields

    table = client.create_table(table)
    return table


def create_bigquery_snapshot(
    project_id: str,
    source_dataset_id: str,
    source_table_name: str,
    destination_dataset_id: str,
    destination_table_name: str,
):
    """Create a BigQuery snapshot from an existing table inside the same project.

    :param project_id: Google Cloud project id.
    :param source_dataset_id: The BigQuery source dataset id.
    :param source_table_name: The BigQuery source table name.
    :param destination_dataset_id: The BigQuery destination dataset id.
    :param destination_table_name: The BigQuery destination table name.
    :return: The table instance if the request was successful.
    """
    func_name = create_bigquery_snapshot.__name__
    msg = (
        f"source: dataset_id={source_dataset_id}, table={source_table_name}, destination: dataset_id="
        f"{destination_dataset_id}, table={destination_table_name}"
    )
    logging.info(f"{func_name}: creating bigquery snapshot {msg}")

    client = bigquery.Client()

    source_dataset_ref = f"{project_id}.{source_dataset_id}"
    source_dataset = bigquery.Dataset(source_dataset_ref)
    source_table = bigquery.Table(source_dataset.table(source_table_name))

    destination_dataset = create_bigquery_dataset(project_id, destination_dataset_id, source_dataset.location)
    destination_table = bigquery.Table(destination_dataset.table(destination_table_name))

    job_config = CopyJobConfig(operation_type="SNAPSHOT", write_disposition="WRITE_EMPTY")
    copy_job: CopyJob = client.copy_table(sources=source_table, destination=destination_table, job_config=job_config)

    copy_job.result()
    success = copy_job.done()
    logging.info(f"{func_name}: create bigquery snapshot {msg}: {success}")
    return success


def load_bigquery_table(
    uri: str,
    dataset_id: str,
    location: str,
    table: str,
    schema_file_path: str,
    source_format: str,
    csv_field_delimiter: str = ",",
    csv_quote_character: str = '"',
    csv_allow_quoted_newlines: bool = False,
    csv_skip_leading_rows: int = 0,
    partition: bool = False,
    partition_field: Union[None, str] = None,
    partition_type: bigquery.TimePartitioningType = bigquery.TimePartitioningType.DAY,
    require_partition_filter=False,
    write_disposition: str = bigquery.WriteDisposition.WRITE_TRUNCATE,
    table_description: str = "",
    project_id: str = None,
    cluster: bool = False,
    clustering_fields=None,
    ignore_unknown_values: bool = False,
) -> bool:
    """Load a BigQuery table from an object on Google Cloud Storage.

    :param uri: the uri of the object to load from Google Cloud Storage into BigQuery.
    :param dataset_id: BigQuery dataset id.
    :param location: location of the BigQuery dataset.
    :param table: BigQuery table name.
    :param schema_file_path: path on local file system to BigQuery table schema.
    :param source_format: the format of the data to load into BigQuery.
    :param csv_field_delimiter: the field delimiter character for data in CSV format.
    :param csv_quote_character: the quote character for data in CSV format.
    :param csv_allow_quoted_newlines: whether to allow quoted newlines for data in CSV format.
    :param csv_skip_leading_rows: the number of leading rows to skip for data in CSV format.
    :param partition: whether to partition the table.
    :param partition_field: the name of the partition field.
    :param partition_type: the type of partitioning.
    :param require_partition_filter: whether the partition filter is required or not when querying the table.
    :param write_disposition: whether to append, overwrite or throw an error when data already exists in the table.
    :param table_description: the description of the table.
    :param project_id: Google Cloud project id.
    :param cluster: whether to cluster the table or not.
    :param clustering_fields: what fields to cluster on.
    Default is to overwrite.
    :param ignore_unknown_values: whether to ignore unknown values or not.
    :return:
    """

    assert uri.startswith("gs://"), "load_big_query_table: 'uri' must begin with 'gs://'"

    func_name = load_bigquery_table.__name__
    msg = (
        f"uri={uri}, project_id={project_id}, dataset_id={dataset_id}, location={location}, table={table}, "
        f"schema_file_path={schema_file_path}, source_format={source_format}"
    )
    logging.info(f"{func_name}: load bigquery table {msg}")

    client = bigquery.Client()
    if project_id is None:
        project_id = client.project
    dataset = bigquery.Dataset(f"{project_id}.{dataset_id}")

    # Handle mutable default arguments
    if clustering_fields is None:
        clustering_fields = []

    # Create load job
    job_config = LoadJobConfig()

    # Set global options
    job_config.source_format = source_format
    job_config.schema = client.schema_from_json(schema_file_path)
    job_config.write_disposition = write_disposition
    job_config.destination_table_description = table_description
    job_config.ignore_unknown_values = ignore_unknown_values

    # Set CSV options
    if source_format == SourceFormat.CSV:
        job_config.field_delimiter = csv_field_delimiter
        job_config.quote_character = csv_quote_character
        job_config.allow_quoted_newlines = csv_allow_quoted_newlines
        job_config.skip_leading_rows = csv_skip_leading_rows

    # Set partitioning settings
    if partition:
        job_config.time_partitioning = bigquery.TimePartitioning(
            type_=partition_type, field=partition_field, require_partition_filter=require_partition_filter
        )
    # Set clustering settings
    if cluster:
        job_config.clustering_fields = clustering_fields

    load_job = None
    try:
        load_job: [LoadJob, None] = client.load_table_from_uri(
            uri, dataset.table(table), location=location, job_config=job_config
        )

        result = load_job.result()
        state = result.state == "DONE"

        logging.info(f"{func_name}: load bigquery table result.state={result.state}, {msg}")
    except BadRequest as e:
        logging.error(f"{func_name}: load bigquery table failed: {e}.")
        if load_job:
            logging.error(f"Error collection:\n{load_job.errors}")
        state = False

    return state


def run_bigquery_query(query: str, bytes_budget: Optional[int] = 549755813888) -> list:
    """Run a BigQuery query.  Defaults to 0.5 TiB query budget.

    :param query: the query to run.
    :param bytes_budget: Maximum bytes allowed to be processed by the query.
    :return: the results.
    """

    bytes_estimate = bq_query_bytes_estimate(query)

    bq_query_bytes_budget_check(bytes_budget=bytes_budget, bytes_estimate=bytes_estimate)

    bq_query_bytes_daily_limit_check(bytes_estimate)

    client = bigquery.Client()
    query_job = client.query(query)
    rows = query_job.result()
    success = query_job.errors is None

    return list(rows)


def copy_bigquery_table(
    source_table_id: Union[str, list],
    destination_table_id: str,
    data_location: str,
    write_disposition: bigquery.WriteDisposition = bigquery.WriteDisposition.WRITE_TRUNCATE,
) -> bool:
    """Copy a BigQuery table.

    :param source_table_id: the id of the source table, including the project name and dataset id.
    :param destination_table_id: the id of the destination table, including the project name and dataset id.
    :param data_location: the location of the datasets.
    :param write_disposition: whether to append, overwrite or throw an error when data already exists in the table.
    :return: whether the table was copied successfully or not.
    """
    func_name = copy_bigquery_table.__name__
    msg = f"source_table_ids={source_table_id}, destination_table_id={destination_table_id}, location={data_location}"
    logging.info(f"{func_name}: copying bigquery table {msg}")

    client = bigquery.Client()
    job_config = bigquery.CopyJobConfig()

    job_config.write_disposition = write_disposition

    job = client.copy_table(source_table_id, destination_table_id, location=data_location, job_config=job_config)
    result = job.result()
    return result.done()


def create_bigquery_view(project_id: str, dataset_id: str, view_name: str, query: str) -> None:
    """Create a BigQuery view.

    :param project_id: the Google Cloud project id.
    :param dataset_id: the BigQuery dataset id.
    :param view_name: the name to call the view.
    :param query: the query for the view.
    :return: None
    """

    client = bigquery.Client()
    dataset = bigquery.DatasetReference(project_id, dataset_id)
    view_ref = dataset.table(view_name)
    view = bigquery.Table(view_ref)
    view.view_query = query
    view = client.create_table(view, exists_ok=True)


def create_bigquery_table_from_query(
    sql: str,
    project_id: str,
    dataset_id: str,
    table_id: str,
    location: str,
    labels=None,
    query_parameters=None,
    clustering_fields=None,
    bytes_budget: Optional[int] = 549755813888,
    schema_file_path: str = None,
) -> bool:
    """Create a BigQuery dataset from a provided query. Defaults to 0.5 TiB query budget.
    If a schema file path is given and the table does not exist yet, then an empty table will be created with this
    schema

    :param sql: the sql query to be executed
    :param labels: labels to place on the new table
    :param project_id: the Google Cloud project id
    :param dataset_id: the BigQuery dataset id
    :param table_id: the BigQuery table id
    :param location: the location where the dataset will be stored:
    https://cloud.google.com/compute/docs/regions-zones/#locations
    :param query_parameters: parameters for a parametrised query.
    :param clustering_fields: what fields to cluster on.
    :param bytes_budget: Maximum bytes allowed to be processed by query.
    :param schema_file_path: path on local file system to BigQuery table schema.
    :return:
    """

    # Handle mutable default arguments
    if labels is None:
        labels = {}
    if query_parameters is None:
        query_parameters = []

    func_name = create_bigquery_table_from_query.__name__
    msg = (
        f"project_id={project_id}, dataset_id={dataset_id}, location={location}, table={table_id}, "
        f"schema_file_path(optional)={schema_file_path}"
    )
    logging.info(f"{func_name}: create bigquery table from query, {msg}")

    # Make bigquery client handle
    client = bigquery.Client()

    dataset = bigquery.DatasetReference(project_id, dataset_id)
    dataset.location = location
    table = bigquery.Table(dataset.table(table_id))

    write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE

    # Create empty table with schema if given and table does not exist yet, change write disposition to empty so schema
    # is not overwritten
    if schema_file_path:
        try:
            client.get_table(table)
        except NotFound:
            table = create_empty_bigquery_table(
                dataset_id,
                location,
                table_id,
                schema_file_path,
                project_id=project_id,
                clustering_fields=clustering_fields,
            )
            write_disposition = bigquery.WriteDisposition.WRITE_EMPTY

    job_config = bigquery.QueryJobConfig(
        allow_large_results=True,
        destination=table,
        labels=labels,
        use_legacy_sql=False,
        query_parameters=query_parameters,
        write_disposition=write_disposition,
    )

    if clustering_fields:
        job_config.clustering_fields = clustering_fields

    bytes_estimate = bq_query_bytes_estimate(sql, job_config=job_config)

    bq_query_bytes_budget_check(bytes_budget=bytes_budget, bytes_estimate=bytes_estimate)

    bq_query_bytes_daily_limit_check(bytes_estimate)

    query_job: QueryJob = client.query(sql, job_config=job_config)
    query_job.result()
    success = query_job.done()
    logging.info(f"{func_name}: create bigquery table from query {msg}: {success}")
    return success


def storage_bucket_exists(bucket_name: str):
    """Check whether the Google Cloud Storage bucket exists

    :param bucket_name: Bucket name (without gs:// prefix)
    :return: Whether the bucket exists or not
    """
    client = storage.Client()
    bucket = client.bucket(bucket_name)

    exists = bucket.exists()

    return exists


def create_cloud_storage_bucket(
    bucket_name: str, location: str = None, project_id: str = None, lifecycle_delete_age: int = None
) -> bool:
    """Create a cloud storage bucket

    :param bucket_name: The name of the bucket. Bucket names must start and end with a number or letter.
    :param location: (Optional) The location of the bucket. If not passed, the default location, US, will be used.
    :param project_id: The project which the client acts on behalf of. Will be passed when creating a topic. If not
    passed, falls back to the default inferred from the environment.
    :param lifecycle_delete_age: Days until files in bucket are deleted
    :return: Whether creating bucket was successful or not.
    """
    func_name = create_cloud_storage_bucket.__name__
    logging.info(f"{func_name}: {bucket_name}")

    success = False

    client = storage.Client(project=project_id)
    bucket = storage.Bucket(client, name=bucket_name)
    if lifecycle_delete_age:
        bucket.add_lifecycle_delete_rule(age=lifecycle_delete_age)
    try:
        client.create_bucket(bucket, location=location)
        success = True
    except Conflict:
        logging.info(
            f"{func_name}: bucket already exists, name: {bucket.name}, project: {bucket.project_number}, "
            f"location: {bucket.location}"
        )
    return success


def copy_blob_from_cloud_storage(
    blob_name: str, bucket_name: str, destination_bucket_name: str, new_name: str = None
) -> bool:
    """Copy a blob from one bucket to another

    :param blob_name: The name of the blob. This corresponds to the unique path of the object in the bucket.
    :param bucket_name: The bucket to which the blob belongs.
    :param destination_bucket_name: The bucket into which the blob should be copied.
    :param new_name:  (Optional) The new name for the copied file.
    :return: Whether copy was successful.
    """
    func_name = copy_blob_from_cloud_storage.__name__
    logging.info(f"{func_name}: {os.path.join(bucket_name, blob_name)}")

    client = storage.Client()

    # source blob and bucket
    bucket = storage.Bucket(client, name=bucket_name)
    blob = bucket.blob(blob_name)

    # destination bucket
    destination_bucket = storage.Bucket(client, name=destination_bucket_name)

    # copy
    response = bucket.copy_blob(blob, destination_bucket=destination_bucket, new_name=new_name)
    success = True if response else False

    return success


def download_blob_from_cloud_storage(
    bucket_name: str,
    blob_name: str,
    file_path: str,
    retries: int = 3,
    connection_sem: BoundedSemaphore = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
) -> bool:
    """Download a blob to a file.

    :param bucket_name: the name of the Google Cloud storage bucket.
    :param blob_name: the path to the blob.
    :param file_path: the file path where the blob should be saved.
    :param retries: the number of times to retry downloading the blob.
    :param connection_sem: a BoundedSemaphore to limit the number of download connections that can run at once.
    :param chunk_size: the chunk size to use when downloading a blob in multiple parts, must be a multiple of 256 KB.
    :return: whether the download was successful or not.
    """

    func_name = download_blob_from_cloud_storage.__name__
    logging.info(f"{func_name}: {file_path}")

    # Get blob
    client = storage.Client()
    bucket = client.bucket(bucket_name)
    blob: Blob = bucket.blob(blob_name)

    # State
    download = True
    success = False

    # Get hash for blob, have to call reload
    blob.reload()
    expected_hash = blob.crc32c

    # Check if file exists and check hash
    if os.path.exists(file_path):
        # Check file's hash
        logging.info(f"{func_name}: file exists, checking hash {file_path}")
        actual_hash = crc32c_base64_hash(file_path)

        # Compare hashes
        files_match = expected_hash == actual_hash
        logging.info(
            f"{func_name}: files_match={files_match}, expected_hash={expected_hash}, " f"actual_hash={actual_hash}"
        )
        if files_match:
            logging.info(
                f"{func_name}: skipping download as files match bucket_name={bucket_name}, "
                f"blob_name={blob_name}, file_path={file_path}"
            )
            download = False
            success = True

    if download:
        # Get connection semaphore
        if connection_sem is not None:
            connection_sem.acquire()

        for i in range(0, retries):
            try:
                blob.chunk_size = chunk_size
                blob.download_to_filename(file_path)
                success = True
                break
            except ChunkedEncodingError as e:
                logging.error(f"{func_name}: exception downloading file: try={i}, file_path={file_path}, exception={e}")

        # Release connection semaphore
        if connection_sem is not None:
            connection_sem.release()

    return success


def download_blobs_from_cloud_storage(
    bucket_name: str,
    prefix: str,
    destination_path: str,
    max_processes: int = cpu_count(),
    max_connections: int = cpu_count(),
    retries: int = 3,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
) -> bool:
    """Download all blobs on a Google Cloud Storage bucket that are within a prefixed path, to a destination on the
    local file system.

    :param bucket_name: the name of the Google Cloud storage bucket.
    :param prefix: the prefixed path on the bucket, where blobs will be searched for.
    :param destination_path: the destination on the local file system to download files too.
    :param max_processes: the maximum number of processes.
    :param max_connections: the maximum number of download connections at once.
    :param retries: the number of times to retry downloading the blob.
    :param chunk_size: the chunk size to use when downloading a blob in multiple parts, must be a multiple of 256 KB.
    :return: whether the files were downloaded successfully or not.
    """

    func_name = download_blobs_from_cloud_storage.__name__

    # Get bucket
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(bucket_name)

    # List blobs
    blobs: List[Blob] = list(bucket.list_blobs(prefix=prefix))
    logging.info(f"{func_name}: {blobs}")

    # Download each blob in parallel
    manager = multiprocessing.Manager()
    connection_sem = manager.BoundedSemaphore(value=max_connections)
    with ProcessPoolExecutor(max_workers=max_processes) as executor:
        # Create tasks
        futures = []
        futures_msgs = {}
        for blob in blobs:
            # Save files to destination path, remove blobs_path from blob name
            filename = f'{os.path.normpath(destination_path)}{blob.name.replace(prefix, "")}'
            msg = f"bucket_name={bucket_name}, blob_name={blob.name}, filename={filename}"
            logging.info(f"{func_name}: {msg}")

            # Create directory
            dirname = os.path.dirname(filename)
            os.makedirs(dirname, exist_ok=True)

            future = executor.submit(
                download_blob_from_cloud_storage,
                bucket_name,
                blob.name,
                filename,
                retries=retries,
                connection_sem=connection_sem,
                chunk_size=chunk_size,
            )
            futures.append(future)
            futures_msgs[future] = msg

        # Wait for completed tasks
        results = []
        for future in as_completed(futures):
            success = future.result()
            results.append(success)
            msg = futures_msgs[future]

            if success:
                logging.info(f"{func_name}: success, {msg}")
            else:
                logging.info(f"{func_name}: failed, {msg}")

    return all(results)


def upload_files_to_cloud_storage(
    bucket_name: str,
    blob_names: List[str],
    file_paths: List[str],
    max_processes: int = cpu_count(),
    max_connections: int = cpu_count(),
    retries: int = 3,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
) -> bool:
    """Upload a list of files to Google Cloud storage.

    :param bucket_name: the name of the Google Cloud storage bucket.
    :param blob_names: the destination paths of blobs where the files will be uploaded.
    :param file_paths: the paths of the files to upload as blobs.
    :param max_processes: the maximum number of processes.
    :param max_connections: the maximum number of upload connections at once.
    :param retries: the number of times to retry uploading a file if an error occurs.
    :param chunk_size: the chunk size to use when uploading a blob in multiple parts, must be a multiple of 256 KB.
    :return: whether the files were uploaded successfully or not.
    """

    func_name = upload_files_to_cloud_storage.__name__
    logging.info(f"{func_name}: uploading files")

    # Upload each file in parallel
    manager = multiprocessing.Manager()
    connection_sem = manager.BoundedSemaphore(value=max_connections)
    with ProcessPoolExecutor(max_workers=max_processes) as executor:
        # Create tasks
        futures = []
        futures_msgs = {}
        for blob_name, file_path in zip(blob_names, file_paths):
            msg = f"{func_name}: bucket_name={bucket_name}, blob_name={blob_name}, file_path={str(file_path)}"
            logging.info(f"{func_name}: {msg}")
            future = executor.submit(
                upload_file_to_cloud_storage,
                bucket_name,
                blob_name,
                file_path=str(file_path),
                retries=retries,
                connection_sem=connection_sem,
                chunk_size=chunk_size,
            )
            futures.append(future)
            futures_msgs[future] = msg

        # Wait for completed tasks
        results = []
        for future in as_completed(futures):
            print(f"future type is {type(future)}")
            success, upload = future.result()
            results.append(success)
            msg = futures_msgs[future]
            if success:
                logging.info(f"{func_name}: success, {msg}")
            else:
                logging.info(f"{func_name}: failed, {msg}")

    return all(results)


def upload_file_to_cloud_storage(
    bucket_name: str,
    blob_name: str,
    file_path: str,
    retries: int = 3,
    connection_sem: BoundedSemaphore = None,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    project_id: str = None,
    check_blob_hash: bool = True,
) -> Tuple[bool, bool]:
    """Upload a file to Google Cloud Storage.

    :param bucket_name: the name of the Google Cloud Storage bucket.
    :param blob_name: the name of the blob to save.
    :param file_path: the path of the file to upload.
    :param retries: the number of times to retry uploading a file if an error occurs.
    :param connection_sem: a BoundedSemaphore to limit the number of upload connections that can run at once.
    :param chunk_size: the chunk size to use when uploading a blob in multiple parts, must be a multiple of 256 KB.
    :param project_id: the project in which the bucket is located, defaults to inferred from the environment.
    :param check_blob_hash: check whether the blob exists and if the crc32c hashes match, in which case skip uploading.
    :return: whether the task was successful or not and whether the file was uploaded.
    """
    func_name = upload_file_to_cloud_storage.__name__
    logging.info(f"{func_name}: bucket_name={bucket_name}, blob_name={blob_name}")

    # State
    upload = True
    success = False

    # Get blob
    storage_client = storage.Client(project=project_id)
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)

    # Check if blob exists already and matches the file we are uploading
    if check_blob_hash and blob.exists():
        # Get blob hash
        blob.reload()
        expected_hash = blob.crc32c

        # Check file hash
        actual_hash = crc32c_base64_hash(file_path)

        # Compare hashes
        files_match = expected_hash == actual_hash
        logging.info(
            f"{func_name}: files_match={files_match}, expected_hash={expected_hash}, " f"actual_hash={actual_hash}"
        )
        if files_match:
            logging.info(
                f"{func_name}: skipping upload as files match. bucket_name={bucket_name}, blob_name={blob_name}, "
                f"file_path={file_path}"
            )
            upload = False
            success = True

    # Upload if file doesn't exist or exists and doesn't match
    if upload:
        # Get connection semaphore
        if connection_sem is not None:
            connection_sem.acquire()

        for i in range(0, retries):
            try:
                blob.chunk_size = chunk_size
                blob.upload_from_filename(file_path)
                success = True
                break
            except ChunkedEncodingError as e:
                logging.error(f"{func_name}: exception uploading file: try={i}, exception={e}")

        # Release connection semaphore
        if connection_sem is not None:
            connection_sem.release()

    return success, upload


def google_cloud_storage_transfer_job(job: dict, func_name: str, gc_project_id: str) -> Tuple[bool, int]:
    """Start a google cloud storage transfer job

    :param job: contains the details of the transfer job
    :param func_name: function name used for detailed logging info
    :param gc_project_id: the Google Cloud project id that holds the Google Cloud Storage bucket.
    :return: whether the transfer was a success or not and the number of objects transferred.
    """
    client = gcp_api.build("storagetransfer", "v1")
    create_result = client.transferJobs().create(body=job).execute()
    transfer_job_name = create_result["name"]

    transfer_job_filter = json.dumps({"project_id": gc_project_id, "job_names": [transfer_job_name]})
    wait_time = 60
    while True:
        response = client.transferOperations().list(name="transferOperations", filter=transfer_job_filter).execute()
        if "operations" in response:
            operations = response["operations"]

            in_progress_count, success_count, failed_count, aborted_count, objects_count = (0, 0, 0, 0, 0)
            for op in operations:
                status = op["metadata"]["status"]
                if status == TransferStatus.success.value:
                    success_count += 1
                elif status == TransferStatus.failed.value:
                    failed_count += 1
                elif status == TransferStatus.aborted.value:
                    aborted_count += 1
                elif status == TransferStatus.in_progress.value:
                    in_progress_count += 1

                try:
                    objects_found = int(op["metadata"]["counters"]["objectsFoundFromSource"])
                except KeyError:
                    objects_found = 0
                objects_count += objects_found

            num_operations = len(operations)
            logging.info(
                f"{func_name}: transfer job {transfer_job_name} operations success={success_count}, "
                f"failed={failed_count}, aborted={aborted_count}, in_progress_count={in_progress_count}, "
                f"objects_count={objects_count}"
            )

            if success_count >= num_operations:
                status = TransferStatus.success
                break
            elif failed_count >= 1:
                status = TransferStatus.failed
                break
            elif aborted_count >= 1:
                status = TransferStatus.aborted
                break

        time.sleep(wait_time)

    return status == TransferStatus.success, objects_count


def azure_to_google_cloud_storage_transfer(
    *,
    azure_storage_account_name: str,
    azure_sas_token: str,
    azure_container: str,
    include_prefixes: List[str],
    gc_project_id: str,
    gc_bucket: str,
    description: str,
    gc_bucket_path: str = None,
    start_date: pendulum.DateTime = pendulum.now("UTC"),
) -> bool:
    """Transfer files from an Azure blob container to a Google Cloud Storage bucket.

    :param azure_storage_account_name: the name of the Azure Storage account that holds the Azure blob container.
    :param azure_sas_token: the shared access signature (SAS) for the Azure blob container.
    :param azure_container: the name of the Azure Blob container where files will be copied from.
    :param include_prefixes: the prefixes of blobs to download from the Azure blob container.
    :param gc_project_id: the Google Cloud project id that holds the Google Cloud Storage bucket.
    :param gc_bucket: the Google Cloud bucket name.
    :param description: a description for the transfer job.
    :param gc_bucket_path: the path in the Google Cloud bucket to save the objects.
    :param start_date: the date that the transfer job will start.
    :return: whether the transfer was a success or not.
    """

    # Leave out scheduleStartTime to make sure that the job starts immediately
    # Make sure to specify scheduleEndDate as today otherwise the job will repeat
    func_name = azure_to_google_cloud_storage_transfer.__name__

    job = {
        "description": description,
        "status": "ENABLED",
        "projectId": gc_project_id,
        "schedule": {
            "scheduleStartDate": {"day": start_date.day, "month": start_date.month, "year": start_date.year},
            "scheduleEndDate": {"day": start_date.day, "month": start_date.month, "year": start_date.year},
        },
        "transferSpec": {
            "azureBlobStorageDataSource": {
                "storageAccount": azure_storage_account_name,
                "azureCredentials": {"sasToken": azure_sas_token},
                "container": azure_container,
            },
            "objectConditions": {"includePrefixes": include_prefixes},
            "gcsDataSink": {"bucketName": gc_bucket},
        },
    }

    if gc_bucket_path is not None:
        # Must end in a / see https://cloud.google.com/storage-transfer/docs/create-manage-transfer-program
        if not gc_bucket_path.endswith("/"):
            gc_bucket_path = f"{gc_bucket_path}/"

        job["transferSpec"]["gcsDataSink"]["path"] = gc_bucket_path

    success, objects_count = google_cloud_storage_transfer_job(job, func_name, gc_project_id)
    return success


def aws_to_google_cloud_storage_transfer(
    aws_access_key_id: str,
    aws_secret_key: str,
    aws_bucket: str,
    include_prefixes: List[str],
    gc_project_id: str,
    gc_bucket: str,
    description: str,
    gc_bucket_path: str = None,
    last_modified_since: pendulum.DateTime = None,
    last_modified_before: pendulum.DateTime = None,
    transfer_manifest: str = None,
    start_date: pendulum.DateTime = pendulum.now("UTC"),
) -> Tuple[bool, int]:
    """Transfer files from an AWS bucket to a Google Cloud Storage bucket.

    :param aws_access_key_id: the id of the key for the aws S3 bucket.
    :param aws_secret_key: the secret key for the aws S3 bucket.
    :param aws_bucket: the name of the aws S3 bucket where files will be copied from.
    :param include_prefixes: the prefixes of blobs to download from the Azure blob container.
    :param gc_project_id: the Google Cloud project id that holds the Google Cloud Storage bucket.
    :param gc_bucket: the Google Cloud bucket name.
    :param description: a description for the transfer job.
    :param gc_bucket_path: the destination folder inside the Google Cloud bucket.
    :param last_modified_since:
    :param last_modified_before:
    :param transfer_manifest: Path to manifest file in Google Cloud bucket (incl gs://)
    :param start_date: the date that the transfer job will start.
    :return: whether the transfer was a success or not.
    """

    # Leave out scheduleStartTime to make sure that the job starts immediately
    # Make sure to specify scheduleEndDate as today otherwise the job will repeat
    func_name = aws_to_google_cloud_storage_transfer.__name__

    job = {
        "description": description,
        "status": "ENABLED",
        "projectId": gc_project_id,
        "schedule": {
            "scheduleStartDate": {"day": start_date.day, "month": start_date.month, "year": start_date.year},
            "scheduleEndDate": {"day": start_date.day, "month": start_date.month, "year": start_date.year},
        },
        "transferSpec": {
            "awsS3DataSource": {
                "bucketName": aws_bucket,
                "awsAccessKey": {"accessKeyId": aws_access_key_id, "secretAccessKey": aws_secret_key},
            },
            "objectConditions": {"includePrefixes": include_prefixes},
            "gcsDataSink": {"bucketName": gc_bucket},
        },
    }
    if gc_bucket_path:
        job["transferSpec"]["gcsDataSink"]["path"] = gc_bucket_path

    if last_modified_since:
        job["transferSpec"]["objectConditions"]["lastModifiedSince"] = last_modified_since.isoformat().replace(
            "+00:00", "Z"
        )
    if last_modified_before:
        job["transferSpec"]["objectConditions"]["lastModifiedBefore"] = last_modified_before.isoformat().replace(
            "+00:00", "Z"
        )

    if transfer_manifest:
        job["transferSpec"]["transferManifest"] = {"location": transfer_manifest}

    success, objects_count = google_cloud_storage_transfer_job(job, func_name, gc_project_id)
    return success, objects_count


def select_table_shard_dates(
    project_id: str, dataset_id: str, table_id: str, end_date: Union[pendulum.DateTime, pendulum.Date], limit: int = 1
) -> List[pendulum.Date]:
    """Returns a list of table shard dates, sorted from the most recent to the oldest date. By default it returns
    the first result.
    :param project_id: the Google Cloud project id.
    :param dataset_id: the BigQuery dataset id.
    :param table_id: the table id (without the date suffix on the end).
    :param end_date: the end date of the table suffixes to search for (most recent date).
    :param limit: the number of results to return.
    :return:
    """

    template_path = os.path.join(utils_templates_path(), make_sql_jinja2_filename("select_table_shard_dates"))
    query = render_template(
        template_path,
        project_id=project_id,
        dataset_id=dataset_id,
        table_id=table_id,
        end_date=end_date.strftime("%Y-%m-%d"),
        limit=limit,
    )
    rows = run_bigquery_query(query)
    dates = []
    for row in rows:
        py_date = row["suffix"]
        date = pendulum.Date(py_date.year, py_date.month, py_date.day)
        dates.append(date)
    return dates


def bq_delete_old_rows(
    project_id: str, dataset_id: str, table_id: str, identifier: str, bytes_budget: Optional[int] = None
):
    """Delete older entries from an ingestion time partitioned table.
    When there are multiple rows with the same identifier, but a different partition date the rows with the older
    partition date(s) are deleted.
    Note that if there are multiple rows with the same identifier and same partition date, that all of those rows
    will be deleted.

    :param project_id: GCP project id.
    :param dataset_id: BigQuery dataset id
    :param table_id: BigQuery table id
    :param identifier: Field in the table on which newer/older rows are matched
    :param bytes_budget: Maximum bytes allowed to be processed.
    :return: None.
    """
    logging.info(f"Deleting old data from table based on partition date")
    template_path = os.path.join(utils_templates_path(), make_sql_jinja2_filename("delete_older_partition"))
    query = render_template(
        template_path,
        project_id=project_id,
        dataset=dataset_id,
        table=table_id,
        identifier=identifier,
    )
    run_bigquery_query(query, bytes_budget=bytes_budget)


def delete_bucket_dir(*, bucket_name: str, prefix: str):
    """Recursively delete blobs from a GCS bucket with a folder prefix.

    :param bucket_name: Bucket name.
    :param prefix: Directory prefix.
    """

    client = storage.Client()
    bucket = client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=prefix)

    for blob in blobs:
        blob.delete()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Sources:
# * https://github.com/apache/airflow/blob/ffb472cf9e630bd70f51b74b0d0ea4ab98635572/airflow/cli/commands/task_command.py
# * https://github.com/apache/airflow/blob/master/docs/apache-airflow/best-practices.rst

# Copyright (c) 2011-2017 Ruslan Spivak
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Sources:
# * https://github.com/rspivak/sftpserver/blob/master/src/sftpserver/__init__.py

# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

import contextlib
import datetime
import logging
import os
import shutil
import socket
import socketserver
import threading
import time
import unittest
import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from multiprocessing import Process
from typing import Dict, List
from unittest.mock import patch

import croniter
import google
import httpretty
import paramiko
import pendulum
import requests
from airflow import DAG, settings
from airflow.exceptions import AirflowException
from airflow.models import DagBag
from airflow.models.connection import Connection
from airflow.models.dagrun import DagRun
from airflow.models.taskinstance import TaskInstance
from airflow.models.variable import Variable
from airflow.operators.dummy_operator import DummyOperator
from airflow.utils import db
from airflow.utils.state import State
from click.testing import CliRunner
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from dateutil.relativedelta import relativedelta
from freezegun import freeze_time
from google.cloud import bigquery, storage
from google.cloud.exceptions import NotFound
from observatory.api.testing import ObservatoryApiEnvironment
from observatory.platform.elastic.elastic_environment import ElasticEnvironment
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.file_utils import (
    crc32c_base64_hash,
    get_file_hash,
    gzip_file_crc,
    list_to_jsonl_gz,
)
from observatory.platform.utils.gc_utils import (
    SourceFormat,
    bigquery_sharded_table_id,
    load_bigquery_table,
    upload_files_to_cloud_storage,
)
from observatory.platform.utils.workflow_utils import find_schema
from pendulum import DateTime
from sftpserver.stub_sftp import StubServer, StubSFTPServer
import freezegun


def random_id():
    """Generate a random id for bucket name.

    :return: a random string id.
    """
    return str(uuid.uuid4()).replace("-", "")


def test_fixtures_path(*subdirs) -> str:
    """Get the path to the Observatory Platform test data directory.

    :return: he Observatory Platform test data directory.
    """

    base_path = module_file_path("tests.fixtures")
    return os.path.join(base_path, *subdirs)


def find_free_port(host: str = "localhost") -> int:
    """Find a free port.

    :param host: the host.
    :return: the free port number
    """

    with socketserver.TCPServer((host, 0), None) as tcp_server:
        return tcp_server.server_address[1]


def save_empty_file(path: str, file_name: str) -> str:
    """Save empty file and return path.

    :param path: the file directory.
    :param file_name: the file name.
    :return: the full file path.
    """

    file_path = os.path.join(path, file_name)
    open(file_path, "a").close()

    return file_path


class ObservatoryEnvironment:
    OBSERVATORY_HOME_KEY = "OBSERVATORY_HOME"

    def __init__(
        self,
        project_id: str = None,
        data_location: str = None,
        api_host: str = "localhost",
        api_port: int = 5000,
        enable_api: bool = True,
        enable_elastic: bool = False,
        elastic_port: int = 9200,
        kibana_port: int = 5601,
    ):
        """Constructor for an Observatory environment.

        To create an Observatory environment:
        env = ObservatoryEnvironment()
        with env.create():
            pass

        :param project_id: the Google Cloud project id.
        :param data_location: the Google Cloud data location.
        :param api_host: the Observatory API host.
        :param api_port: the Observatory API port.
        :param enable_api: whether to enable the observatory API or not.
        :param enable_elastic: whether to enable the Elasticsearch and Kibana test services.
        :param elastic_port: the Elastic port.
        :param kibana_port: the Kibana port.
        """

        self.project_id = project_id
        self.data_location = data_location
        self.api_host = api_host
        self.api_port = api_port
        self.buckets = []
        self.datasets = []
        self.data_path = None
        self.session = None
        self.temp_dir = None
        self.api_env = None
        self.api_session = None
        self.enable_api = enable_api
        self.enable_elastic = enable_elastic
        self.elastic_port = elastic_port
        self.kibana_port = kibana_port
        self.dag_run: DagRun = None
        self.elastic_env: ElasticEnvironment = None

        if self.create_gcp_env:
            self.download_bucket = self.add_bucket()
            self.transform_bucket = self.add_bucket()
            self.storage_client = storage.Client()
            self.bigquery_client = bigquery.Client()
        else:
            self.download_bucket = None
            self.transform_bucket = None
            self.storage_client = None
            self.bigquery_client = None

    @property
    def create_gcp_env(self) -> bool:
        """Whether to create the Google Cloud project environment.

        :return: whether to create Google Cloud project environ,ent
        """

        return self.project_id is not None and self.data_location is not None

    def assert_gcp_dependencies(self):
        """Assert that the Google Cloud project dependencies are met.

        :return: None.
        """

        assert self.create_gcp_env, "Please specify the Google Cloud project_id and data_location"

    def add_bucket(self) -> str:
        """Add a Google Cloud Storage Bucket to the Observatory environment.

        The bucket will be created when create() is called and deleted when the Observatory
        environment is closed.

        :return: returns the bucket name.
        """

        self.assert_gcp_dependencies()
        bucket_name = random_id()
        self.buckets.append(bucket_name)
        return bucket_name

    def _create_bucket(self, bucket_id: str) -> None:
        """Create a Google Cloud Storage Bucket.

        :param bucket_id: the bucket identifier.
        :return: None.
        """

        self.assert_gcp_dependencies()
        self.storage_client.create_bucket(bucket_id, location=self.data_location)

    def _create_dataset(self, dataset_id: str) -> None:
        """Create a BigQuery dataset.

        :param dataset_id: the dataset identifier.
        :return: None.
        """

        self.assert_gcp_dependencies()
        dataset = bigquery.Dataset(f"{self.project_id}.{dataset_id}")
        dataset.location = self.data_location
        self.bigquery_client.create_dataset(dataset, exists_ok=True)

    def _delete_bucket(self, bucket_id: str) -> None:
        """Delete a Google Cloud Storage Bucket.

        :param bucket_id: the bucket identifier.
        :return: None.
        """

        self.assert_gcp_dependencies()
        try:
            bucket = self.storage_client.get_bucket(bucket_id)
            bucket.delete(force=True)
        except requests.exceptions.ReadTimeout:
            pass
        except google.api_core.exceptions.NotFound:
            logging.warning(
                f"Bucket {bucket_id} not found. Did you mean to call _delete_bucket on the same bucket twice?"
            )

    def add_dataset(self, prefix: str = "") -> str:
        """Add a BigQuery dataset to the Observatory environment.

        The BigQuery dataset will be deleted when the Observatory environment is closed.

        :param prefix: an optional prefix for the dataset.
        :return: the BigQuery dataset identifier.
        """

        self.assert_gcp_dependencies()
        if prefix != "":
            dataset_id = f"{prefix}_{random_id()}"
        else:
            dataset_id = random_id()
        self.datasets.append(dataset_id)
        return dataset_id

    def _delete_dataset(self, dataset_id: str) -> None:
        """Delete a BigQuery dataset.

        :param dataset_id: the BigQuery dataset identifier.
        :return: None.
        """

        self.assert_gcp_dependencies()
        try:
            self.bigquery_client.delete_dataset(dataset_id, not_found_ok=True, delete_contents=True)
        except requests.exceptions.ReadTimeout:
            pass

    def add_variable(self, var: Variable) -> None:
        """Add an Airflow variable to the Observatory environment.

        :param var: the Airflow variable.
        :return: None.
        """

        self.session.add(var)
        self.session.commit()

    def add_connection(self, conn: Connection):
        """Add an Airflow connection to the Observatory environment.

        :param conn: the Airflow connection.
        :return: None.
        """

        self.session.add(conn)
        self.session.commit()

    def run_task(self, task_id: str) -> TaskInstance:
        """Run an Airflow task.

        :param task_id: the Airflow task identifier.
        :return: None.
        """

        assert self.dag_run is not None, "with create_dag_run must be called before run_task"

        dag = self.dag_run.dag
        run_id = self.dag_run.run_id
        task = dag.get_task(task_id=task_id)
        ti = TaskInstance(task, run_id=run_id)
        ti.dag_run = self.dag_run
        ti.init_run_context(raw=True)
        ti.run(ignore_ti_state=True)

        return ti

    @contextlib.contextmanager
    def create_dag_run(self, dag: DAG, execution_date: pendulum.DateTime, freeze: bool = True):
        """Create a DagRun that can be used when running tasks.
        During cleanup the DAG run state is updated.

        :param dag: the Airflow DAG instance.
        :param execution_date: the execution date of the DAG.
        :param freeze: whether to freeze time to the start date of the DAG run.
        :return: None.
        """
        # Get start date, which is one schedule interval after execution date
        if isinstance(dag.normalized_schedule_interval, (timedelta, relativedelta)):
            start_date = (
                datetime.fromtimestamp(execution_date.timestamp(), pendulum.tz.UTC) + dag.normalized_schedule_interval
            )
        else:
            start_date = croniter.croniter(dag.normalized_schedule_interval, execution_date).get_next(pendulum.DateTime)

        # Disable freezegun for the observatory api client. OpenAPI client generated code does not like FakeDateTime
        freezegun.configure(extend_ignore_list=["observatory.api.client"])
        frozen_time = freeze_time(start_date, tick=True)

        run_id = "manual__{0}".format(execution_date.isoformat())

        # Make sure google auth uses real DateTime and not freezegun fake time
        with patch("google.auth._helpers.utcnow", wraps=datetime.utcnow) as mock_utc_now:
            try:
                if freeze:
                    frozen_time.start()
                state = State.RUNNING
                self.dag_run = dag.create_dagrun(
                    run_id=run_id,
                    state=state,
                    execution_date=execution_date,
                    start_date=pendulum.now("UTC"),
                )
                yield self.dag_run
            finally:
                self.dag_run.update_state()
                if freeze:
                    frozen_time.stop()

    @contextlib.contextmanager
    def create(self, task_logging: bool = False):
        """Make and destroy an Observatory isolated environment, which involves:

        * Creating a temporary directory.
        * Setting the OBSERVATORY_HOME environment variable.
        * Initialising a temporary Airflow database.
        * Creating download and transform Google Cloud Storage buckets.
        * Creating default Airflow Variables: AirflowVars.DATA_PATH, AirflowVars.PROJECT_ID, AirflowVars.DATA_LOCATION,
          AirflowVars.DOWNLOAD_BUCKET and AirflowVars.TRANSFORM_BUCKET.
        * Cleaning up all resources when the environment is closed.

        :param task_logging: display airflow task logging
        :yield: Observatory environment temporary directory.
        """

        with CliRunner().isolated_filesystem() as temp_dir:
            # Set temporary directory
            self.temp_dir = temp_dir

            # Prepare environment
            self.new_env = {self.OBSERVATORY_HOME_KEY: os.path.join(self.temp_dir, ".observatory")}
            prev_env = dict(os.environ)

            try:
                # Update environment
                os.environ.update(self.new_env)

                # Create Airflow SQLite database
                settings.DAGS_FOLDER = os.path.join(self.temp_dir, "airflow", "dags")
                os.makedirs(settings.DAGS_FOLDER, exist_ok=True)
                airflow_db_path = os.path.join(self.temp_dir, "airflow.db")
                settings.SQL_ALCHEMY_CONN = f"sqlite:///{airflow_db_path}"
                logging.info(f"SQL_ALCHEMY_CONN: {settings.SQL_ALCHEMY_CONN}")
                settings.configure_orm(disable_connection_pool=True)
                self.session = settings.Session
                db.initdb()

                # Setup Airflow task logging
                original_log_level = logging.getLogger().getEffectiveLevel()
                if task_logging:
                    # Set root logger to INFO level, it seems that custom 'logging.info()' statements inside a task
                    # come from root
                    logging.getLogger().setLevel(20)
                    # Propagate logging so it is displayed
                    logging.getLogger("airflow.task").propagate = True

                # Create buckets and datasets
                if self.create_gcp_env:
                    for bucket_id in self.buckets:
                        self._create_bucket(bucket_id)

                    for dataset_id in self.datasets:
                        self._create_dataset(dataset_id)

                # Add default Airflow variables
                self.data_path = os.path.join(self.temp_dir, "data")
                self.add_variable(Variable(key=AirflowVars.DATA_PATH, val=self.data_path))

                # Add Google Cloud environment related Airflow variables
                if self.create_gcp_env:
                    self.add_variable(Variable(key=AirflowVars.PROJECT_ID, val=self.project_id))
                    self.add_variable(Variable(key=AirflowVars.DATA_LOCATION, val=self.data_location))
                    self.add_variable(Variable(key=AirflowVars.DOWNLOAD_BUCKET, val=self.download_bucket))
                    self.add_variable(Variable(key=AirflowVars.TRANSFORM_BUCKET, val=self.transform_bucket))

                # Start elastic
                if self.enable_elastic:
                    elastic_build_path = os.path.join(self.temp_dir, "elastic")
                    self.elastic_env = ElasticEnvironment(
                        build_path=elastic_build_path, elastic_port=self.elastic_port, kibana_port=self.kibana_port
                    )
                    self.elastic_env.start()
                self.dag_run: DagRun = None

                # Create ObservatoryApiEnvironment
                if self.enable_api:
                    self.api_env = ObservatoryApiEnvironment(host=self.api_host, port=self.api_port)
                    with self.api_env.create():
                        self.api_session = self.api_env.session
                        yield self.temp_dir
                else:
                    yield self.temp_dir
            finally:
                # Set logger settings back to original settings
                logging.getLogger().setLevel(original_log_level)
                logging.getLogger("airflow.task").propagate = False

                # Revert environment
                os.environ.clear()
                os.environ.update(prev_env)

                if self.create_gcp_env:
                    # Remove Google Cloud Storage buckets
                    for bucket_id in self.buckets:
                        self._delete_bucket(bucket_id)

                    # Remove BigQuery datasets
                    for dataset_id in self.datasets:
                        self._delete_dataset(dataset_id)

                # Stop elastic
                if self.enable_elastic:
                    self.elastic_env.stop()


class ObservatoryTestCase(unittest.TestCase):
    """Common test functions for testing Observatory Platform DAGs"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(ObservatoryTestCase, self).__init__(*args, **kwargs)
        self.storage_client = storage.Client()
        self.bigquery_client = bigquery.Client()

        # Turn logging to warning because vcr prints too much at info level
        logging.basicConfig()
        vcr_log = logging.getLogger("vcr")
        vcr_log.setLevel(logging.WARNING)

    def assert_dag_structure(self, expected: Dict, dag: DAG):
        """Assert the DAG structure.

        :param expected: a dictionary of DAG task ids as keys and values which should be a list of downstream task ids.
        :param dag: the DAG.
        :return: None.
        """

        expected_keys = expected.keys()
        actual_keys = dag.task_dict.keys()
        self.assertEqual(expected_keys, actual_keys)

        for task_id, downstream_list in expected.items():
            self.assertTrue(dag.has_task(task_id))
            task = dag.get_task(task_id)
            self.assertEqual(set(downstream_list), task.downstream_task_ids)

    def assert_dag_load(self, dag_id: str, dag_file: str):
        """Assert that the given DAG loads from a DagBag.

        :param dag_id: the DAG id.
        :param dag_file: the path to the DAG file.
        :return: None.
        """

        with CliRunner().isolated_filesystem() as dag_folder:
            if not os.path.exists(dag_file):
                raise Exception(f"{dag_file} does not exist.")

            shutil.copy(dag_file, os.path.join(dag_folder, os.path.basename(dag_file)))

            dag_bag = DagBag(dag_folder=dag_folder)

            if dag_bag.import_errors != {}:
                logging.error(f"DagBag errors: {dag_bag.import_errors}")
            self.assertEqual({}, dag_bag.import_errors)

            dag = dag_bag.get_dag(dag_id=dag_id)

            if dag is None:
                logging.error(
                    f"DAG not found in the database. Make sure the DAG ID is correct, and the dag file contains the words 'airflow' and 'DAG'."
                )
            self.assertIsNotNone(dag)

            self.assertGreaterEqual(len(dag.tasks), 1)

    def assert_blob_exists(self, bucket_id: str, blob_name: str):
        """Assert whether a blob exists or not.

        :param bucket_id: the Google Cloud storage bucket id.
        :param blob_name: the blob name (full path except for bucket)
        :return: None.
        """

        # Get blob
        bucket = self.storage_client.get_bucket(bucket_id)
        blob = bucket.blob(blob_name)
        self.assertTrue(blob.exists())

    def assert_blob_integrity(self, bucket_id: str, blob_name: str, local_file_path: str):
        """Assert whether the blob uploaded and that it has the expected hash.

        :param blob_name: the Google Cloud Blob name, i.e. the entire path to the blob on the Cloud Storage bucket.
        :param bucket_id: the Google Cloud Storage bucket id.
        :param local_file_path: the path to the local file.
        :return: whether the blob uploaded and that it has the expected hash.
        """

        # Get blob
        bucket = self.storage_client.get_bucket(bucket_id)
        blob = bucket.blob(blob_name)
        result = blob.exists()

        # Check that blob hash matches if it exists
        if result:
            # Get blob hash
            blob.reload()
            expected_hash = blob.crc32c

            # Check actual file
            actual_hash = crc32c_base64_hash(local_file_path)
            result = expected_hash == actual_hash

        self.assertTrue(result)

    def assert_table_integrity(self, table_id: str, expected_rows: int = None):
        """Assert whether a BigQuery table exists and has the expected number of rows.

        :param table_id: the BigQuery table id.
        :param expected_rows: the expected number of rows.
        :return: whether the table exists and has the expected number of rows.
        """

        table = None
        actual_rows = None
        try:
            table = self.bigquery_client.get_table(table_id)
            actual_rows = table.num_rows
        except NotFound:
            pass

        self.assertIsNotNone(table)
        if expected_rows is not None:
            self.assertEqual(expected_rows, actual_rows)

    def assert_table_content(self, table_id: str, expected_content: List[dict] = None):
        """Assert whether a BigQuery table has any content and if expected content is given whether it matches the
        actual content. The order of the rows is not checked, only whether all rows in the expected content match
        the rows in the actual content.
        The expected content should be a list of dictionaries, where each dictionary represents one row of the table,
        the keys are fieldnames and values are values.

        :param table_id: the BigQuery table id.
        :param expected_content: the expected content.
        :return: whether the table has content and the expected content is correct
        """
        rows = None
        actual_content = None
        try:
            rows = self.bigquery_client.list_rows(table_id)
            actual_content = [dict(row) for row in rows]
        except NotFound:
            pass

        self.assertIsNotNone(rows)
        if expected_content is not None:
            for row in expected_content:
                self.assertIn(row, actual_content)
                actual_content.remove(row)
            self.assertListEqual(
                [], actual_content, msg=f"Rows in actual content that are not in expected content: {actual_content}"
            )

    def assert_table_bytes(self, table_id: str, expected_bytes: int):
        """Assert whether the given bytes from a BigQuery table matches the expected bytes.

        :param table_id: the BigQuery table id.
        :param expected_bytes: the expected number of bytes.
        :return: whether the table exists and the expected bytes match
        """

        table = None
        try:
            table = self.bigquery_client.get_table(table_id)
        except NotFound:
            pass

        self.assertIsNotNone(table)
        self.assertEqual(expected_bytes, table.num_bytes)

    def assert_file_integrity(self, file_path: str, expected_hash: str, algorithm: str):
        """Assert that a file exists and it has the correct hash.

        :param file_path: the path to the file.
        :param expected_hash: the expected hash.
        :param algorithm: the algorithm to use when hashing, either md5 or gzip crc
        :return: None.
        """

        self.assertTrue(os.path.isfile(file_path))

        if algorithm == "gzip_crc":
            actual_hash = gzip_file_crc(file_path)
        else:
            actual_hash = get_file_hash(file_path=file_path, algorithm=algorithm)

        self.assertEqual(expected_hash, actual_hash)

    def assert_cleanup(self, download_folder: str, extract_folder: str, transform_folder: str):
        """Assert that the download, extracted and transformed folders were cleaned up.

        :param download_folder: the path to the DAGs download folder.
        :param extract_folder: the path to the DAGs extract folder.
        :param transform_folder: the path to the DAGs transform folder.
        :return: None.
        """

        self.assertFalse(os.path.exists(download_folder))
        self.assertFalse(os.path.exists(extract_folder))
        self.assertFalse(os.path.exists(transform_folder))

    def setup_mock_file_download(
        self, uri: str, file_path: str, headers: Dict = None, method: str = httpretty.GET
    ) -> None:
        """Use httpretty to mock a file download.

        This function must be called from within an httpretty.enabled() block, for instance:

        with httpretty.enabled():
            self.setup_mock_file_download('https://example.com/file.zip', path_to_file)

        :param uri: the URI of the file download to mock.
        :param file_path: the path to the file on the local system.
        :param headers: the response headers.
        :return: None.
        """

        if headers is None:
            headers = {}

        with open(file_path, "rb") as f:
            body = f.read()

        httpretty.register_uri(method, uri, adding_headers=headers, body=body)


class SftpServer:
    """A Mock SFTP server for testing purposes"""

    def __init__(
        self,
        host: str = "localhost",
        port: int = 3373,
        level: str = "INFO",
        backlog: int = 10,
        startup_wait_secs: int = 1,
        socket_timeout: int = 10,
    ):
        """Create a Mock SftpServer instance.

        :param host: the host name.
        :param port: the port.
        :param level: the log level.
        :param backlog: ?
        :param startup_wait_secs: time in seconds to wait before returning from create to give the server enough
        time to start before connecting to it.
        """

        self.host = host
        self.port = port
        self.level = level
        self.backlog = backlog
        self.startup_wait_secs = startup_wait_secs
        self.is_shutdown = True
        self.tmp_dir = None
        self.root_dir = None
        self.private_key_path = None
        self.server_thread = None
        self.socket_timeout = socket_timeout

    def _generate_key(self):
        """Generate a private key.

        :return: the filepath to the private key.
        """

        key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

        private_key_path = os.path.join(self.tmp_dir, "test_rsa.key")
        with open(private_key_path, "wb") as f:
            f.write(
                key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.TraditionalOpenSSL,
                    encryption_algorithm=serialization.NoEncryption(),
                )
            )

        return private_key_path

    def _start_server(self):
        paramiko_level = getattr(paramiko.common, self.level)
        paramiko.common.logging.basicConfig(level=paramiko_level)

        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.settimeout(self.socket_timeout)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        server_socket.bind((self.host, self.port))
        server_socket.listen(self.backlog)

        while not self.is_shutdown:
            try:
                conn, addr = server_socket.accept()
                transport = paramiko.Transport(conn)
                transport.add_server_key(paramiko.RSAKey.from_private_key_file(self.private_key_path))
                transport.set_subsystem_handler("sftp", paramiko.SFTPServer, StubSFTPServer)

                server = StubServer()
                transport.start_server(server=server)

                channel = transport.accept()
                while transport.is_active() and not self.is_shutdown:
                    time.sleep(1)

            except socket.timeout:
                # Timeout must be set for socket otherwise it will wait for a connection forever and block
                # the thread from exiting. At: conn, addr = server_socket.accept()
                pass

    @contextlib.contextmanager
    def create(self):
        """Make and destroy a test SFTP server.

        :yield: None.
        """

        with CliRunner().isolated_filesystem() as tmp_dir:
            # Override the root directory of the SFTP server, which is set as the cwd at import time
            self.tmp_dir = tmp_dir
            self.root_dir = os.path.join(tmp_dir, "home")
            os.makedirs(self.root_dir, exist_ok=True)
            StubSFTPServer.ROOT = self.root_dir

            # Generate private key
            self.private_key_path = self._generate_key()

            try:
                self.is_shutdown = False
                self.server_thread = threading.Thread(target=self._start_server)
                self.server_thread.start()

                # Wait a little bit to give the server time to grab the socket
                time.sleep(self.startup_wait_secs)

                yield self.root_dir
            finally:
                # Stop server and wait for server thread to join
                self.is_shutdown = True
                if self.server_thread is not None:
                    self.server_thread.join()


def make_dummy_dag(dag_id: str, execution_date: pendulum.DateTime) -> DAG:
    """A Dummy DAG for testing purposes.

    :param dag_id: the DAG id.
    :param execution_date: the DAGs execution date.
    :return: the DAG.
    """

    with DAG(
        dag_id=dag_id,
        schedule_interval="@weekly",
        default_args={"owner": "airflow", "start_date": execution_date},
        catchup=False,
    ) as dag:
        task1 = DummyOperator(task_id="dummy_task")

    return dag


@dataclass
class Table:
    """A table to be loaded into Elasticsearch.

    :param table_name: the table name.
    :param is_sharded: whether the table is sharded or not.
    :param dataset_id: the dataset id.
    :param records: the records to load.
    :param schema_prefix: the schema prefix.
    :param schema_folder: the schema path.
    """

    table_name: str
    is_sharded: bool
    dataset_id: str
    records: List[Dict]
    schema_prefix: str
    schema_folder: str


def bq_load_tables(
    *,
    tables: List[Table],
    bucket_name: str,
    release_date: DateTime,
    data_location: str,
    project_id: str = None,
):
    """Load the fake Observatory Dataset in BigQuery.

    :param tables: the list of tables and records to load.
    :param bucket_name: the Google Cloud Storage bucket name.
    :param release_date: the release date for the observatory dataset.
    :param data_location: the location of the BigQuery dataset.
    :param project_id: GCP project id.
    :return: None.
    """

    with CliRunner().isolated_filesystem() as t:
        files_list = []
        blob_names = []

        # Save to JSONL
        for table in tables:
            blob_name = f"{table.dataset_id}-{table.table_name}.jsonl.gz"
            file_path = os.path.join(t, blob_name)
            list_to_jsonl_gz(file_path, table.records)
            files_list.append(file_path)
            blob_names.append(blob_name)

        # Upload to Google Cloud Storage
        success = upload_files_to_cloud_storage(bucket_name, blob_names, files_list)
        assert success, "Data did not load into BigQuery"

        # Save to BigQuery tables
        for blob_name, table in zip(blob_names, tables):
            if table.is_sharded:
                table_id = bigquery_sharded_table_id(table.table_name, release_date)
            else:
                table_id = table.table_name

            # Select schema file based on release date
            schema_file_path = find_schema(table.schema_folder, table.schema_prefix, release_date)
            if schema_file_path is None:
                logging.error(
                    f"No schema found with search parameters: analysis_schema_path={table.schema_folder}, "
                    f"table_name={table.table_name}, release_date={release_date}"
                )
                exit(os.EX_CONFIG)

            # Load BigQuery table
            uri = f"gs://{bucket_name}/{blob_name}"
            logging.info(f"URI: {uri}")
            success = load_bigquery_table(
                uri,
                table.dataset_id,
                data_location,
                table_id,
                schema_file_path,
                SourceFormat.NEWLINE_DELIMITED_JSON,
                project_id=project_id,
            )
            if not success:
                raise AirflowException("bq_load task: data failed to load data into BigQuery")


class HttpServer:
    """Simple HTTP server for testing. Serves files from a directory to http://locahost:port/filename"""

    def __init__(self, directory: str):
        """Initialises the server.

        :param directory: Directory to serve.
        """

        self.directory = directory
        self.process = None

        self.host = "localhost"
        self.port = find_free_port(host=self.host)
        self.address = (self.host, self.port)
        self.url = f"http://{self.host}:{self.port}/"

    @staticmethod
    def serve_(address, directory):
        """Entry point for a new process to run HTTP server.

        :param address: Address (host, port) to bind server to.
        :param directory: Directory to serve.
        """

        os.chdir(directory)
        server = ThreadingHTTPServer(address, SimpleHTTPRequestHandler)
        server.serve_forever()

    def start(self):
        """Spin the server up in a new process."""

        # Don't try to start it twice.
        if self.process is not None and self.process.is_alive():
            return

        self.process = Process(
            target=HttpServer.serve_,
            args=(
                self.address,
                self.directory,
            ),
        )
        self.process.start()

    def stop(self):
        """Shutdown the server."""

        if self.process is not None and self.process.is_alive():
            self.process.kill()
            self.process.join()

    @contextlib.contextmanager
    def create(self):
        """Spin up a server for the duration of the session."""
        self.start()

        try:
            yield self.process
        finally:
            self.stop()



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs

import os

import click

from observatory.platform.cli.cli_utils import indent, INDENT1, INDENT2
from observatory.platform.observatory_config import TerraformConfig
from observatory.platform.terraform.terraform_api import TerraformApi
from observatory.platform.terraform.terraform_api import TerraformVariable
from observatory.platform.terraform.terraform_builder import TerraformBuilder


class TerraformCommand:
    def __init__(self, config: TerraformConfig, terraform_credentials_path: str, debug: bool = False):
        """Create a TerraformCommand, which can be used to create and update terraform workspaces.

        :param config: the Terraform Config file.
        :param terraform_credentials_path: the path to the Terraform credentials file.
        :param debug: whether to print debugging information.
        """

        self.config = config
        self.terraform_builder = TerraformBuilder(config, debug=debug)
        self.terraform_credentials_path = terraform_credentials_path
        self.debug = debug
        self.terraform_credentials_exists = os.path.exists(terraform_credentials_path)

    @property
    def is_environment_valid(self):
        """Whether is the parameters passed to the TerraformCommand are valid.

        :return: whether the parameters passed to the TerraformCommand are valid.
        """

        return all(
            [
                self.terraform_credentials_exists,
                self.terraform_builder.is_environment_valid,
            ]
        )

    def print_variable(self, var: TerraformVariable):
        """Print the output for the CLI for a single TerraformVariable instance.

        :param var: the TerraformVariable instance.
        :return: None.
        """

        if var.sensitive:
            print(indent(f"* \x1B[3m{var.key}\x1B[23m: sensitive", INDENT2))
        else:
            print(indent(f"* \x1B[3m{var.key}\x1B[23m: {var.value}", INDENT2))

    def print_variable_update(self, old_var: TerraformVariable, new_var: TerraformVariable):
        """Print the output for the CLI for a terraform variable that is being updated.

        :param old_var: the old TerraformVariable instance.
        :param new_var: the new TerraformVariable instance.
        :return: None.
        """

        if old_var.sensitive:
            print(indent(f"* \x1B[3m{old_var.key}\x1B[23m: sensitive -> sensitive", INDENT2))
        else:
            print(indent(f"* \x1B[3m{old_var.key}\x1B[23m: {old_var.value} -> {new_var.value}", INDENT2))

    def build_terraform(self):
        """Build the Terraform files for the Observatory Platform.

        :return: None.
        """

        self.terraform_builder.build_terraform()

    def build_image(self):
        """Build a Google Compute image for the Terraform deployment with Packer.

        :return: None.
        """

        self.terraform_builder.build_image()

    def build_google_container_image(self):
        """Build a Docker image stored in the Google Container registry using gcloud builds.

        :return: None.
        """
        self.terraform_builder.gcloud_activate_service_account()
        self.terraform_builder.gcloud_builds_submit()

    @property
    def verbosity(self):
        """Convert debug switch into Terraform API verbosity.
        :return:
        """

        if self.debug:
            return TerraformApi.VERBOSITY_DEBUG
        return TerraformApi.VERBOSITY_WARNING

    def print_summary(self):
        # Get organization, environment and prefix
        organization = self.config.terraform.organization
        environment = self.config.backend.environment.value
        workspace = self.config.terraform_workspace_id

        # Display settings for workspace
        print("\nTerraform Cloud Workspace: ")
        print(indent(f"Organization: {organization}", INDENT1))
        print(
            indent(
                f"- Name: {workspace} (prefix: '{TerraformConfig.WORKSPACE_PREFIX}' + suffix: '{environment}')", INDENT1
            )
        )
        print(indent(f"- Settings: ", INDENT1))
        print(indent(f"- Auto apply: True", INDENT2))
        print(indent(f"- Terraform Variables:", INDENT1))

    def create_workspace(self):
        """Create a Terraform workspace.

        :return: None.
        """

        self.print_summary()

        # Get terraform token
        token = TerraformApi.token_from_file(self.terraform_credentials_path)
        terraform_api = TerraformApi(token, self.verbosity)

        # Get variables
        terraform_variables = self.config.terraform_variables()

        # Get organization, environment and prefix
        organization = self.config.terraform.organization
        workspace = self.config.terraform_workspace_id

        for variable in terraform_variables:
            self.print_variable(variable)

        # confirm creating workspace
        if click.confirm("Would you like to create a new workspace with these settings?"):
            print("Creating workspace...")

            # Create new workspace
            terraform_api.create_workspace(organization, workspace, auto_apply=True, description="")

            # Get workspace ID
            workspace_id = terraform_api.workspace_id(organization, workspace)

            # Add variables to workspace
            for var in terraform_variables:
                terraform_api.add_workspace_variable(var, workspace_id)

            print("Successfully created workspace")

    def update_workspace(self):
        """Update a Terraform workspace.

        :return: None.
        """

        self.print_summary()

        # Get terraform token
        token = TerraformApi.token_from_file(self.terraform_credentials_path)
        terraform_api = TerraformApi(token, self.verbosity)

        # Get variables
        terraform_variables = self.config.terraform_variables()

        # Get organization, environment and prefix
        organization = self.config.terraform.organization
        workspace = self.config.terraform_workspace_id

        # Get workspace ID
        workspace_id = terraform_api.workspace_id(organization, workspace)
        add, edit, unchanged, delete = terraform_api.plan_variable_changes(terraform_variables, workspace_id)

        if add:
            print(indent("NEW", INDENT1))
            for var in add:
                self.print_variable(var)
        if edit:
            print(indent("UPDATE", INDENT1))
            for old_var, new_var in edit:
                self.print_variable_update(old_var, new_var)
        if delete:
            print(indent("DELETE", INDENT1))
            for var in delete:
                self.print_variable(var)
        if unchanged:
            print(indent("UNCHANGED", INDENT1))
            for var in unchanged:
                self.print_variable(var)

        # confirm creating workspace
        if click.confirm("Would you like to update the workspace with these settings?"):
            print("Updating workspace...")

            # Update variables in workspace
            terraform_api.update_workspace_variables(add, edit, delete, workspace_id)

            print("Successfully updated workspace")
